/*
  PROGRAM TO CREATE OF FILE OF  THE CRC'S OF THE FILES ON THE DEFAULT DISK  

bugs...
	should accept as an argument a file listing the specific files and/or
	directories to check.

  This program was written by Ted H. Emigh, and has been placed in the public
  domain, to be used at the user's discretion.  The CRC routines and the
  discussion of the CRC were written by David Dantowitz, Digital Equipment
  Corporation,  Dantowitz%eagle1.dec@decwrl.

  This program calculates the CRC (cyclic redundancy check) for all the files
  on the disk (with the exception of files that are hidden system files).  The
  CRC's are placed in a file (CHECK$$$.NEW) to be compared with the CRC's
  calculated at a previous time in the file CHECK$$$.CRC.  The comparison is
  done with the program COMPARE.PAS.  This program is set to automatically
  chain to COMPARE.PAS to automate the procedure, but this can be turned off
  by deleting the lines:
    Assign (chain_file, 'COMPARE.CHN');
    Chain(chain_file);
  at the end of this program.


   For a good discussion of polynomial selection see "Cyclic
   Codes for Error Detection", by W. W. Peterson and
   D. T. Brown, Proceedings of the IEEE, volume 49, pp 228-235, 
   January 1961.

   A reference on table driven CRC computation is "A Cyclic
   Redundancy Checking (CRC) Algorithm" by A. B. Marton and
   T. K. Frambs, The Honeywell Computer Journal, volume 5, 
   number 3, 1971.

   Also used to prepare these examples was "Computer Networks", 
   by Andrew S. Tanenbaum, Prentice Hall, Inc.  Englewood Cliffs, 
   New Jersey, 1981.

   The following three polynomials are international standards:


        CRC-12 = X^12 + X^11 + X^3 + X^2 + X^1 + 1
        CRC-16 = X^16 + X^15 + X^2 + 1
        CRC-CCITT = X^16 + X^12 + X^5 + 1

   The polynomials can be represented by a binary number, where a 1
   indicates the inclusion of the power term in the polynomial.  Since
   the highest order term is always included, that term is not needed
   in specifying the polynomial, and usually is dropped.  In addition, 
   the bits are specified from low-order to high-order.  For example, 
   the polynomial CRC-12 can be represented in the following manner:

    Order                   0  1  2  3  4  5  6  7  8  9 10 11 12
    Term Included ?         Y  Y  Y  Y  N  N  N  N  N  N  N  Y  Y
    Binary Representation   1  1  1  1  0  0  0  0  0  0  0  1 (1)<-- DROPPED

   The binary and hex representations for the three polynomials are:

                   Binary                     Hex

        CRC-12    = 1111 0000 0001           0x0F01
        CRC-16    = 1010 0000 0000 0001      0xA001    (Used below)
        CRC-CCITT = 1000 0100 0000 1000      0x8404

   The first is used with 6-bit characters and the second two with
   8-bit characters.  All of the above will detect any odd number of
   errors.  The second two will catch all 16-bit bursts, a high
   percentage of random 17-bit bursts (~99.997%) and also a large
   percentage of random 18-bit or larger bursts (~99.998%).  The paper
   mentioned above (Peterson and Brown) discusses how to compute the
   statistics presented which have been quoted from Tanenbaum.  Notice
   that some errors can be generated in nonrandom ways that can
   substantially reduce the chances of detecting errors.

   (A burst of length N is defined a sequence of N bits, where the
   first and last bits are incorrect and the bits in the middle are any
   possible combination of correct and incorrect.  See the paper by
   Peterson and Brown for more information)

	Modifications by Jim Van Zandt, November 1988...
		Traps control-break, so file attributes aren't left changed.
		Program can be customized at run time.

	Modifications by Jim Van Zandt, August 1988...
		Translated from Pascal to C.
		Adapted assembly language code to DeSmet C compiler.
		Calculating CRC starting at an offset into the file.
		Checking own object code before proceeding.
		Writing time and date of CRC check to output file.
		Implemented options for checking executable, system, hidden
			and/or read-only files.
		Drive to be checked need not be current drive.


	Implementation notes...
		To use a compiler other than DeSmet, substitute the C code for
		the assembly language stuff in array_crc() and rework the calls
		to interrupt 21, done here by function _doint().

	Notes on customizing...

		This program can be customized at run time so that a user need
		not have access to a C compiler in order to change the
		parameters.  The parameters are all in the structure
		"signature".  The program merely searches in its own object
		code for the two strings signature.flag1 and signature.flag2,
		which bracket the adjustable parameters.

		It is conceivable that a program (virus, worm, Trojan horse,
		etc.) could itself execute this program in order to change the
		parameters to something known.  That would, for example, let it
		modify other files without changing their CRCs.  I have tried
		to make this difficult.  The primary defense is that the name
		of the program can be changed.  In addition, the flag strings
		are modified each time the program is customized.  There is no
		fixed string which marks the parameters.

		A user who can compile the program may want to preclude
		run-time customizing by removing the #define of CUSTOMIZABLE,
		below.

*/
#define DESMET
#define CUSTOMIZABLE  /* #define to enable run-time customizing */

/*	These #defines should be customized by each user  */

#define DEFAULT_OFFSET_DISTANCE 9000L    /* some nonnegative LONG constant */
#define FLAG1 "O4^@6CVNB~654f|jh*&"      /* 19 character strings to mark */
#define FLAG2 "%^*%U wTIUHG/&^%&_)"      /* the CRC in the .EXE file */
#define CRC_PROGRAM_NAME "FILECRC.EXE"


/*	customize these names, then change COMPARE to match */

#define COMPARE_PROGRAM_NAME "COMPARE.EXE"
#define CHECK_NEW "CHECK$$$.NEW"
#define CHECK_CRC "CHECK$$$.CRC"

#include	<stdio.h>


#define 	VERSION		"2.10"
#define 	VERSION_DATE		"19 November 1988"

#define 	POLY		0xA001          /*  CRC Polynomial Used  */

#define SECTORS 192 /* Number of 128 byte sectors in the CRC buffer */

#define BUFFER_LENGTH  128*SECTORS /* Length of the CRC buffer */
char byte_string[BUFFER_LENGTH];   /* buffer for calculating CRC */

#ifdef DESMET

#pragma ex+					/* enable #asm directive */
extern unsigned _rax, _rbx, _rcx, _rdx, _res, _rds;

#else

#include	<string.h>		/* not needed with DeSmet C */

#endif  /* DESMET */


							/* evaluates TRUE (nonzero) for equal strings */
#define streq(a, b) (strcmp((a),(b)) == 0)


/*  DTA as used by MSDOS  */
typedef struct {
    char dos[21];
    char attribute;		            /*  Attribute byte  */
    int time_of_day;				/*  Time of Day of File Creation  */
    int date;						/*  Date of File Creation  */
    long file_size;					/*  Size of the File  */
    char filename[13];				/*  ASCIIZ File Name  */
    char junk[85];
    } DTA_RECORD;
/*  Variables used in Calculating the CRC  */
int str_length; 
int recsread; 
int table_256[256]; 		/*CRC Table to speed computations*/

int conservative=0;		/* nonzero if changes to hidden, system, or read-only 
							files are to be displayed on the screen
							(passed on to COMPARE.EXE) */
int executable_files=0;		/* nonzero if checking executable files */
int system_files=0;			/* nonzero if checking system files */
int readonly_files=0;		/* nonzero if checking read only files */
int hidden_files=0;			/* nonzero if checking hidden files */
int all_files=1;			/* nonzero if checking all files */
int control_break=0;		/* set nonzero when user types control-break */
int user_customizing = 0;	/* nonzero when customizing the parameters */
int comparing_crc_files=0;	/* nonzero if a CRC file already existed */

/*  Variables used in setting up the input and output files  */
FILE *filvar; 
FILE *chain_file; 
FILE *outfile; 

char root[256];   /*  Contains the default drive and root directory */

typedef struct
	{char flag1[20];
	unsigned exefile_crc;
	long offset_distance;
	char crc_program[80];
	char compare_program[80];
	char new_crc_file[80];
	char cur_crc_file[80];
	char flag2[20];
	} PARMS;
extern PARMS signature;

unsigned current_date(), current_time();
long lseek();

/*
    This routine computes the remainder values of 0 through 255 divided
  by the polynomial represented by POLY.  These values are placed in a
  table and used to compute the CRC of a block of data efficiently.
  More space is used, but the CRC computation will be faster.



    This implementation only permits polynomials up to degree 16.
*/

void generate_table_256(poly)
int poly; 
{
int val, i, result; 

   for (val = 0; val <= 255; ++val)
      {
      result = val;
      for (i = 1; i <= 8; ++i)
         {
         if (result & 1)
            result = (result >> 1) ^ poly;
         else 
            result >>= 1;
         }
      table_256[val] = result;
      }
}


int file_crc(sfile, offset) FILE *sfile; long offset;
{
	int n, crc1=0;
	long needed;
	long length;
#define min(i,j) ((i)<(j)?(i):(j))

	length=lseek(sfile, 0L, 2);
/*printf("   file_crc: offset %ld, file length %ld bytes", offset, length);*/
	if(length==0) return crc1;
	offset = offset%length;
	lseek(sfile, offset, 0);
	while(n=fread(byte_string, 1, BUFFER_LENGTH, sfile))
		{
/*printf("read %u bytes\n", n);*/
		crc1=array_crc(byte_string, n, crc1);
		}
/*printf("completed tail of file\n");*/
	lseek(sfile, 0L, 0);
	needed=offset;
	while(needed)
		{n=fread(byte_string, 1, (unsigned)min(needed, BUFFER_LENGTH), sfile);
/*printf("read %u bytes\n", n);*/
		crc1=array_crc(byte_string, n, crc1);
		needed -= n;
		}
/*printf("completed beginning of file\n");*/
/*	printf(",  crc=%04x\n", crc1);*/
	return crc1;
}
	
/*
	crc16 - calculate 16 bit cyclic redundancy check (CRC) word

	This algorithm yields the same values as the one used by ARC and
	PKARC when used with the polynomial CRC-16.  It is a translation
	into C of the program crck4.asm from the info-ibmpc library.
*/

int array_crc(s, len, crc_value)
int crc_value;	/* starting value (0 for the first byte) */
char *s;
int len;
{
/*----------------------------------------------------------------------------
				Ted Emigh's assembly code...

$c4/$7e/<s/                {les di,s[bp]            (es:di points to array)  }
$8b/$46/<initial_crc/      {mov ax,initial_crc[bp]  (initial CRC value)      }
$8b/$4e/<s_length/         {mov cx,s_length[bp]     (count)                  }
$be/table_256/             {mov si,offset table_256 (table address)          }


{ next:  }

$26/$32/$05/               {xor al,es:[di]          CRC = CRC XOR next byte  }
$47/                       {inc di                  (point to next byte)     }

{ intermediate steps, see comments for overall effect }

$31/$db/                   {xor bx,bx               (bx <- 0)                }
$86/$d8/                   {xchg al,bl              (bx <- ax and 0FF)       }
$86/$e0/                   {xchg al,ah              (ax <- ax shr 8)         }
$d1/$e3/                   {shl bx,1                (bx <- bx+bx)            }

$33/$00/                   {xor ax,[bx+si]          CRC = (CRC shr 8) XOR
                                                          table[CRC and 0FF] }

$e2/$f0/                   {loop next               (count <- count -1)      }

$89/$46/<s+4);             {mov s+4[bp],ax          (array_crc := CRC)  }
----------------------------------------------------------------------------*/
#ifdef DESMET

	if(len==0) return crc_value;

#asm
	mov		di,[bp+4]	;s
	mov		cx,[bp+6]	;len
	mov		ax,[bp+8]	;crc_value
	mov		si,offset table_256_
nextbyte:
	xor		al,[di]
	inc		di
	xor		bx,bx		;bx <- 0
	xchg	al,bl		;bx <- ax & 0xff
	xchg	al,ah		;ax <- ax >> 8
	shl		bx,1		;bx <- bx+bx
	xor		ax,[bx+si]
	loop	nextbyte

	mov		[bp+8], ax	;crc_value
#endasm

#else /* not DESMET - use slower but transportable pure C version  */

	while(len--)
		crc_value = (crc_value>>8) ^ table_256[(crc_value&0xff)^*s++];
	return crc_value;

#endif /* DESMET */

	return crc_value;
}



/*
  This procedure computes the CRC for a file.  
*/

void get_crc(current_directory, dtap)
char *current_directory; 
DTA_RECORD *dtap; 
{
   int crc_value; 
   static char pathname[256];

   strcpy(pathname, current_directory);
   strcat(pathname, "\\");
   strcat(pathname, dtap->filename);

   /*  Change the Attribute byte so we can always open it  */
   /*    To save some time, this is only done if the file  */
   /*    Has an attribute other than ARCHIVE              */
   if (dtap->attribute & 0xdf)
      set_attr(0, pathname);

   filvar=fopen(pathname, "r");   /*  Open the file  */

   crc_value = file_crc(filvar, signature.offset_distance);
   fprintf(outfile, "%-15s  %-12s  %2d  %6d  %6d  %6ld  %04x\n", 
           current_directory, 
           dtap->filename, 
           dtap->attribute, 
           dtap->time_of_day, 
           dtap->date, 
           dtap->file_size, 
           crc_value);

   fclose(filvar);
   /*  Restore the correct Attribute Byte  */
   if (dtap->attribute & 0xdf)
      set_attr(dtap->attribute, pathname);
}


/*	
	return current time, in the same format as a directory entry
*/
unsigned current_time()
{
	extern unsigned _rax,_rbx,_rcx,_rdx,_rsi,_rdi,_res,_rds,_doint();
	extern char _carryf,_zerof;
	unsigned time, hours, minutes, seconds;

	_rax=0x2c00;				/* get time */
	_doint(0x21);
	hours=(long)_rcx>>8; minutes=(long)_rcx&63; seconds=(long)_rdx>>8;
	time=((((hours)
			<<6)+minutes)
			<<5)+(seconds>>1);
	return time;
}

/*
	return current date, in the same format as a directory entry 
*/
unsigned current_date()
{
	extern unsigned _rax,_rbx,_rcx,_rdx,_rsi,_rdi,_res,_rds,_doint();
	extern char _carryf,_zerof;
	unsigned date, year, month, day;

	_rax=0x2a00;				/* get date */
	_doint(0x21);
	year=(long)_rcx-1980; month=(long)(_rdx>>8); day=(long)(_rdx&31);
	date=	(((year
			<<4)+month)
			<<5)+day;
	return date;
}



/*

  This routine sets the file attributes.  Uses Function $43 in
  Interrupt $21.

  DOS makes it difficult to open and read certain types files
  (e.g., r/o and files that are both hidden and system).  This
  program gets around that restriction by always setting the
  attribute byte to 0, then resetting it to the original value.

  attr  is the attribute to be set on the file
  asciiz is a string variable with the file name

*/

set_attr(attr, asciiz)
char attr, *asciiz; 
{
   _rax = 0x4301;
   _rds = _showds();
   _rdx = asciiz;
   _rcx = attr;
   _doint(0x21);
}


/*  Function to tell if the file is a directory entry  */

int is_directory(dta)
DTA_RECORD *dta; 
{
	return (dta->attribute & 16) && (dta->filename[0] != '.');
}


/*   sets the disk DTA
    Uses MSDOS Function $1A with interrupt $21
    offset is the offset of the new DTA
    segment is the segment of the new DTA
*/

void set_dta(offset, segment)
int offset, segment; 
{

   _rax = 0x1a00;
   _rds = segment;
   _rdx = offset;
   _doint(0x21);
}


/*   gets the disk DTA
    Uses MSDOS Function $2F with Interrupt $21
    offset will return with the current DTA offset
    segment will return with the current DTA segment
*/

void get_dta(offset, segment)
int *offset, *segment; 
{
   _rax = 0x2f00;
   _doint(0x21);
   *offset = _rbx;
   *segment = _res;
}


/*
    Find the first file matching the ASCIIZ string.
    attr_mask is $27 for files only and $37 for directories & files

    INT 21 function 4EH
    Returns TRUE if found, FALSE if not found
*/

int find_first(attr_mask, asciiz)
unsigned char attr_mask; 
char asciiz[256]; 
{
   _rax = 0x4e00;
   _rds = _showds();
   _rdx = asciiz;
   _rcx = attr_mask;
   _doint(0x21);
   return (_rax != 18);
}


/*
    Find the next file matching the ASCIIZ string.
    attr_mask is $27 for files only and $37 for directories & files

    Returns TRUE if found, FALSE if not found
*/

int find_next(attr_mask)
unsigned char attr_mask; 
{
   _rax = 0x4f00;
   _rcx = attr_mask;
   _doint(0x21);
   return (_rax != 18);
}


/*
  Procedure to calculate the CRC of all the files in a directory, 
  then all subdirectories in that directory

  current_directory contains the directory name (including drive)

*/

void directory(current_directory)  char *current_directory; 
{
int dta_ofs; 
int dta_seg;    /*  Contains the current DTA address  */

DTA_RECORD dta;    /*  Local DTA  */
char this_directory[256]; 
char pathname[256]; 
char asciiz[256];    /* directory and file names */

   get_dta(&dta_ofs, &dta_seg);   /* Save the current DTA location */
   set_dta(&dta, _showds());   /* Set the DTA location to local area */

/*
  Find and print the files in the current directory
*/
   strcpy(asciiz, current_directory);
   strcat(asciiz, "\\*.*");
   /*  Process all the files before doing any directories  */
   if (find_first(0x27, asciiz))
      do
         {
         if (dta.filename[0] != '.')
         	if(all_files || 
         		executable_files && is_executable(&dta) ||
         		hidden_files && is_hidden(&dta) ||
         		system_files && is_system(&dta) ||
         		readonly_files && is_readonly(&dta))

         		get_crc(current_directory, &dta);

         		if(control_break) return;			/* quit gracefully */
         } while (find_next(0x27));

								   /*  Now process all the directories  */
   if (find_first(0x37, asciiz))
      do
         {
         if (is_directory(&dta))
            {
            strcpy(this_directory, current_directory);
            strcat(this_directory, "\\");
            strcat(this_directory, dta.filename);
            printf("%s\n", this_directory);
            			/* recursion is OK here because each 
            				activation has a local DTA */
            directory(this_directory);
            }
         }
      while (find_next(0x37) && !control_break);
   set_dta(dta_ofs, dta_seg);
   /* restore the old DTA */
}

/*
	return nonzero if argument is the DTA for an executable file
*/
int is_executable(dp) DTA_RECORD *dp;
{	
	return strstr(dp->filename, ".COM") ||
	       strstr(dp->filename, ".EXE") ||
	       strstr(dp->filename, ".BAT") ||
	       strstr(dp->filename, ".SYS");		/* remember CONFIG.SYS */
}

/*
	return nonzero if argument is the DTA for a "read only" file
*/
int is_readonly(dp) DTA_RECORD *dp;
{	
	return dp->attribute & 0x01;
}


/*
	return nonzero if argument is the DTA for a "hidden" file
*/
int is_hidden(dp) DTA_RECORD *dp;
{	
	return dp->attribute & 0x02;
}


/*
	return nonzero if argument is the DTA for a "system" file
*/
int is_system(dp) DTA_RECORD *dp;
{	
	return dp->attribute & 0x04;
}


/*
  Function to return the current drive
  Uses MSDOS Function $19 with Interrupt $21
  value returned is 1 for A:, 2 for B:, etc.
*/

unsigned char current_drive()
{
   _rax = 0x1900;
   _doint(0x21);
   return  1 + (0xff & _rax);
}


help()
{
	printf("\nusage:  filecrc  [options]\n");
	printf("options, which may be combined:\n");
	printf("   -c    conservative checking: report changes to \"hidden\", \n");
	printf("         \"system\", or \"read-only\" files to screen\n");
	printf("         (passed on to comparison program)\n");
	printf("   -e    check executable files (.COM, .EXE, .BAT, and .SYS)\n");
	printf("   -h    check \"hidden\" files\n");
	printf("   -r    check \"read-only\" files\n");
	printf("   -s    check \"system\" files\n");
#ifdef CUSTOMIZABLE
	printf("   -u    user customizing - reset CRC offset & other params\n");
#endif
	printf("default: check all files\n");
	exit();
}


break_handler()
{	control_break = 1;		/* this flag is tested in directory() */
	return 0;				/* don't abort execution yet */
}


main(argc, argv) int argc; char *argv[];
{
	char *s, c, **av;
	int ac;

	/*  FILECRC  */
	printf("CRC file integrity program\n");
	printf("Version %s, %s\n", VERSION, VERSION_DATE);
	printf("Written by Ted H. Emigh -- ");
	printf("emigh@ncsugn.uucp or NEMIGH@TUCC.BITNET\n");
	printf("Modified by James R. Van Zandt -- jrv@mitre-bedford.arpa\n");

	ac=argc;
	av=argv;
	argc=1;
	while(--ac)
		{s=*++av;
		if(*s=='?') help();
		if(*s=='-')
			{while(c=*++s)
				{if(c=='?') help();
				else if(c=='c') {conservative=1;}
				else if(c=='e') {executable_files=1; all_files=0;}
				else if(c=='s') {system_files=1; all_files=0;}
				else if(c=='h') {hidden_files=1; all_files=0;}
				else if(c=='r') {readonly_files=1; all_files=0;}
#ifdef CUSTOMIZABLE
				else if(c=='u') {user_customizing = 1;}
#endif
				else printf("invalid command line switch %c ignored\n", c);
				}
			}
		else
			argv[argc++]=s;
		}
							/*  Generate the table for CRC checks  */
	generate_table_256(POLY);    	

							/*	Check the .EXE file for this program */
	check_self();

					/*  initialize root with the current drive designation  */
	root[0]=current_drive()-1+'A';
	root[1]=':';
	root[2]=0;
					/* if a drive was specified on command line, use it */
	if(argc>1)
		{s = argv[1];
		if(s[1] == ':' && 'A'<=toupper(s[0]) && toupper(s[0])<='Z' && s[2]==0)
			root[0] = toupper(s[0]);
		}
							/*  See if CHECK$$$.CRC exists  */
	filvar=fopen(signature.cur_crc_file, "r");
	if (filvar)
	   {
	   fclose(filvar);
	   unlink(signature.new_crc_file);
	   outfile=fopen(signature.new_crc_file,"w");
	   printf("Creating File %s\n", signature.new_crc_file);
	   comparing_crc_files=1;
	   }
	else 
	   {
	   unlink(signature.cur_crc_file);
	   outfile=fopen(signature.cur_crc_file,"w");
	   printf("Creating File %s\n", signature.cur_crc_file);
	   }

	printf("%s\\\n", root);
	fprintf(outfile, "%d %d\n", current_time(), current_date());

#ifdef DESMET
	forward_break(break_handler);	/* prepare for control-break */
#endif

	directory(root);             /*  Now, do the CRC check  */

	fclose(outfile);

	if(control_break)
		{printf("CRC calculation aborted\n");
		exit();
		}
	printf("CRC calculation completed\n\n");

						        /* Now compare this with the previous CRCs  */
	if(comparing_crc_files)
		exec(signature.compare_program, conservative?"-c":"");
}

char filler1[]=";;a;lkdsjf;lzkasjfd;lkja;dlkfja;lkfj;alkjf;ladi";
char filler2[]=";;akdjf;lzksjf;lzkajf;kajs;lvkfa;lkdjfakdjrpaoi";

/*
	based on a parameter update program in DeSmet Gazette, Dec 86
		by King Computer Services, Inc.
	modified for DeSmet environment by Pacific Data Works
	modified to search the path by J. R. Van Zandt
*/
PARMS signature = 
	{FLAG1, 
	0, 
	DEFAULT_OFFSET_DISTANCE, 
	CRC_PROGRAM_NAME, 
	COMPARE_PROGRAM_NAME, 
	CHECK_NEW,
	CHECK_CRC,
	FLAG2
	};

/* check_self - calculate CRC on this program's .EXE file */
check_self()
{	long lseek(), findinst(), fpos, fpos_crc, c;
	FILE *exefile;		/* file to be modified */
	int		flaglen, replacements=0, parmsize, j;
	char *pname();
	PARMS file_parms;
	unsigned calc_crc;

	if(!pname(root, sizeof(root)-1))
		strcpy(root, signature.crc_program);
	printf("\nchecking %s for its CRC \"signature\"\n", root);
	parmsize=sizeof(PARMS);
	flaglen=strlen(signature.flag1)+1;
	while((exefile=open(root,2))==ERR) /* open for binary update */
		{printf("file %s not found\n", root);
		while(1)
			{printf("Indicate... a)bort   r)equest new name   \
c)ontinue without CRC check: ");
			c=tolower(getchar());
			printf("\n");
			if(c == 'a') exit();
			if(c == 'c') return 0;
			if(c == 'r') break;
			}
		printf("type path and filename for this program: ");
		fgets(root, sizeof(root)-1, stdin);
		root[strlen(root) - 1] = 0;		/* zap the \n */
		}

	fpos=findinst(exefile, &signature, signature.flag2, 
		byte_string, BUFFER_LENGTH);
	fpos_crc = fpos + sizeof(signature.flag1);
	lseek(exefile, fpos, 0);
	fread(&file_parms, 1, sizeof(signature), exefile);

/*
	show_parms("parameters read from file", &file_parms);
	show_parms("parameters in program", &signature);
*/

	lseek(exefile, fpos + 1, 0);
	if(findinst(exefile, &signature, signature.flag2, 
		byte_string, BUFFER_LENGTH))
		{/* 
			We've found a second copy of the signature strings.  This
			probably means the compiler left an extra copy of the flag
			strings in some buffer, then included the buffer in the
			compiled code.  To work around the problem, make some
			change to the source code like adding a big string, then
			recompile.  Changing the flag strings might be a good idea,
			too - just in case they match a piece of object code
			somewhere!
		*/
		fprintf(stderr, 
			"more than one CRC in .EXE file - don't know which to use!");
		exit(1);   
		}

	if(file_parms.exefile_crc != signature.exefile_crc)
		{printf("\007WARNING: the CRC signature in the RAM doesn't match the \
one in %s.\n", root);
		printf("Tampering by a virus is the probable cause.\n\n");
		printf("abort execution? (y/n) ");
		c=tolower(getchar());
		if(c!='n') exit();
		}
	else if(signature.exefile_crc == 0)
		{printf("WARNING: The CRC in %s has not \n", root);
		printf("been initialized, or has been tampered with.\n\n");
		printf("   abort execution? (y/n) ");
		c=tolower(getchar());
		if(c!='n') exit();
		printf("\ninitialize the CRC? (y/n) ");
		c=tolower(getchar());
		printf("\n");
		if(c!='n')
			{
/*
	Implementation notes...  

	Rather than recording the CRC for the executable file separately, I
	adjust the contents of one word to force the CRC for the file to a
	particular value.  Suppose the CRC calculated for a block of data
	is X.  If X is appended to the data (added to the end) then the CRC
	calculated for the whole block will be zero.  If X^0xfffe is
	appended instead, then the CRC for the whole block will be 0x2000. 
	(I determined this by experiment.)

	The CRC is initialized to zero at the start of the calculation.  If
	the data consists only of zeros, the calculated CRC will be zero
	regardless of the size of the block.  Therefore, one can add a
	block of zeros to the beginning of the data without changing the
	CRC.  In the following calculation, I really need the CRC for the
	data starting just AFTER signature.exefile_crc, wrapping around to
	the beginning, and ending just BEFORE signature.exefile_crc. 
	However, this would require a function to process "all but 2 bytes"
	in addition to file_crc() which processes all bytes.  Instead, I
	initialize the two bytes of the CRC to zero and start the CRC
	calculation there.
*/
			calc_crc=file_crc(exefile, fpos_crc);
/*			printf("calculated CRC is %04x\n", calc_crc); */
					/* modify so that CRC calculated later will be 0x2000 */
			calc_crc ^= 0xfffe;
			lseek(exefile, fpos_crc, 0);
			fwrite(&calc_crc, 1, 2, exefile);
			printf("the CRC is initialized\n\n");
			printf("   check files too? (y/n) ");
			c=tolower(getchar());
			if(c!='y') exit();
			}
		}
	else
		{
		if(file_crc(exefile, fpos_crc + 2) != 0x2000)
			{printf("\007WARNING: %s fails the CRC test!\n", 
				root);
			printf("The file has been corrupted, or possibly ");
			printf("altered by a virus.  \nabort execution? (y/n) ");
			c=tolower(getchar());
			if(c!='n') exit();
			}
		else
			{printf("CRC test passed\n\n");
#ifdef CUSTOMIZABLE
			if(user_customizing)
				{
				customize(&signature);
				signature.exefile_crc = 0;
/*
				show_parms("parameters read from file", &file_parms);
				show_parms("new parameters", &signature);
*/
				lseek(exefile, fpos, 0);
				fwrite(&signature, 1, sizeof(signature), exefile);
				printf("customized parameters recorded and CRC cleared\n");
				printf("%s will have to be initialized when next run\n", 
														signature.crc_program);
				if(file_parms.offset_distance != signature.offset_distance
										&& fopen(signature.cur_crc_file, "r"))
					printf("since the offset has been changed, \
%s is invalid and should be deleted\n", signature.cur_crc_file);
				fclose(exefile);
				exit(0);
				}
#endif
			}
		}
	fclose(exefile);
}

/*
	This is the generic routine to find a flag string in an .exe file
	to be used as a pointer to the beginning of installable parameters. 
	Returns a long containing the position of the flag, or zero if the
	flag is not found.  File pointer points to the beginning of the
	struct.  Buffer, which must be long enough, contains the structure
	starting at the flag.

	Method...
	1. Read bytes into a buffer at least as long as the struct.
	2. Search the buffer for a match on both flags.
	3. If no match, move data at end of buffer to beginning of buffer,
	   read in another batch from file.
	4. Go to 2 until match found.
	5. When match found, move matched code to head of buffer.
	6. lseek to the file position found and return the fpos.
*/
long findinst(filen, flag1, flag2, buffer, buflen)
FILE *filen;
char *flag1, *flag2, *buffer;
unsigned buflen;
{	long f_pos;
	int nomatch, len;
	int structlen;
	int j, i, istop, k, flaglen;

	nomatch=1;		/* initialize status to "no match" */
	structlen=flag2-flag1+strlen(flag2)+1;

/*
	flaglen=strlen(flag)+1;
	buflen=len=structlen+flaglen;
	j=0;
*/
	f_pos=fseek(filen, 0L, 1);
	len=0;
	while(nomatch)
		{len += read(filen, &buffer[len], buflen-len-1);
		if(len<structlen)	/* if we couldn't read enough bytes */
			return(0L);

		i=0;	/* step through buffer char by char */
		istop=len-structlen+1;

		while(((buffer[i] != *flag1)||
				(nomatch=strcmp(&buffer[i], flag1))||
				(nomatch=strcmp(&buffer[i+flag2-flag1], flag2))) &&
			(i<istop))
			i++;		/* searching for the flags */
		if (nomatch)
			{/*	
				if not found, move the tail end of the data to the head
				of the buffer, set the number of bytes in the buffer,
				reset file pointer, and return for another read
			*/
			f_pos += i;
			for (j=0; i<len;  ) buffer[j++]=buffer[i++];
			len=j;
			}
		else
			{f_pos+=i;		/* update filepointer with the pos of match */
							/* move structure to head of buffer */
			for (j=0; j<structlen;  ) buffer[j++]=buffer[i++];

			}
		}
	lseek(filen, f_pos, 0);	/* set file pointer to beginning of structure */
	return (f_pos);
}


/*
	pname - get pathname of currently running program (else returns 0)
*/
char *pname(buf, max_chars) char *buf; int max_chars;
{	unsigned length, seg,   off;
/*           [bp-2]  [bp-4] [bp-6]  */

#ifdef DESMET

#asm
	mov	ax,3000h	; check if DOS 3.0 or later
	int	21H		; (force al=0 in case of DOS 1)
	cmp	al,3
	jb	argv7		; DOS 1 or 2, return null parameter
	mov	ax,cs
	sub	ax,10h		; calculate PSP segment (for DeSmet compiler)
	mov	es,ax
	mov	es,es:[2ch]	; get environment segment from PSP
	mov	ax,es
	mov	[bp-4],ax	; seg = environment segment
	xor	di,di		; find the program name by
	xor	al,al		; first skipping over all the
	mov	cx,-1		; environment variables...
	cld
argv9:	repne	scasb		; scan for double null (can't use
	scasb			; scasw since might be odd addr.)
	jne	argv9		; loop if it was a single null
	add	di,2		; skip count word in environment
	mov	[bp-6],di	; off = program name offset
	mov	cx,-1		; now find its length...
	repne	scasb		; scan for another null byte
	not	cx		; convert cx to length
	dec	cx
	jmp	argvx
;	REP MOVSB		; es:di++ = ds:si++, while(cx--)

argv7:	xor	cx,cx
argvx:	mov	[bp-2],cx	; length = program name length
#endasm

#else
	return 0;
#endif /* DESMET */

	if(length==0) 
		return 0;
	if(length>max_chars) length=max_chars;
	_lmove(length, off, seg, buf, _showds());
	return buf;
}
	
#ifdef DESMET
/*
	This function redirects the control-break interrupt to the user's
	function handler().  If that function returns a nonzero value, the
	program is terminated with that value used as the return code.  If
	the function returns a zero value, the control-break is ignored
	and control is returned to the function in progress.

	The user's function may execute any DOS function, but must
	preserve bp, cs, ss, and sp.  

	Use with small model only.
*/
forward_break(handler) int handler();
{
#asm
	jmp	forward_1
;
;	note that these are in the code segment, because ds 
;	will be invalid when the break handler gets control
;
handler_addr:	dw	0
ds_copy:	dw	0

;
;	control-break sends control here first
;
brk_forwarder:
	pushf
	push	ds
	push	es
	push	ax
	push	bx
	push	cx
	push	dx
	push	si
	push	di
	mov	ax,ds_copy
	mov	ds,ax		; restore ds
	call	handler_addr	; invoke the user's function

	or	ax,ax		; if he returns a nonzero value, 
	jnz	terminate	; terminate the program
	pop	di
	pop	si
	pop	dx
	pop	cx
	pop	bx
	pop	ax
	pop	es
	pop	ds
	popf
	iret

terminate:
	mov	ah,4cH
	int	21H		; (preserve return code in al, if any)

forward_1:
	mov	ax,ds
	mov	ds_copy, ax	; save copy of ds

	mov	ax, [bp+4]
	mov	word handler_addr, ax	; save address of user's routine

	push	ds
	mov	ax,cs
	mov	ds,ax
	mov	dx,offset brk_forwarder
	mov	ax,2523H
	int	21H		; reset interrupt vector 23 (^C handler)
	pop	ds
#endasm
}

#endif /* DESMET */

#ifdef DEBUGGING_CODE
dump(s, n) char *s; int n;
{	int i, j;
	for (j=0; j<n; j+=16)
		{for (i = j; i < j + 16 && i < n; i++)
			printf("%02x ", s[i]);
		printf("  '");
		for (i = j; i < j + 16 && i < n; i++)
			printf("%c ", isprint(s[i])?s[i]:'.');
		printf("'\n");
		}
}

show_parms(s, pp) char *s; PARMS *pp;
{
	int i;

	printf("%s\n", s);
	printf("          flag1: \"");
	for (i=0; i<20; i++) 
		printf("%c", isprint(pp->flag1[i])?(pp->flag1[i]):'.');
	printf("\"\n            crc: %04x\n", pp->exefile_crc);
	printf("         offset: %ld\n", pp->offset_distance);
	printf("    crc_program: \"");
	for (i=0; i<12; i++) 
		printf("%c", isprint(pp->crc_program[i])?(pp->crc_program[i]):'.');
	printf("\"\n");
	printf("compare_program: \"");
	for (i=0; i<12; i++) 
		printf("%c", isprint(pp->compare_program[i])?(pp->compare_program[i]):'.');
	printf("\"\n");
	printf("   new_crc_file: \"");
	for (i=0; i<12; i++) 
		printf("%c", isprint(pp->new_crc_file[i])?(pp->new_crc_file[i]):'.');
	printf("\"\n");
	printf("   cur_crc_file: \"");
	for (i=0; i<12; i++) 
		printf("%c", isprint(pp->cur_crc_file[i])?(pp->cur_crc_file[i]):'.');
	printf("\"\n");
	printf("          flag2: \"");
	for (i=0; i<20; i++) 
		printf("%c", isprint(pp->flag2[i])?(pp->flag2[i]):'.');
	printf("\"\n");
}
#endif

/*
	customize parameters as instructed by the user


for reference:

typedef struct
	{char flag1[20];
	unsigned exefile_crc;
	long offset_distance;
	char crc_program[80];
	char compare_program[80];
	char new_crc_file[80];
	char cur_crc_file[80];
	char flag2[20];
	} PARMS;
*/

customize(pp) PARMS *pp;
{	long ll, atol();
	int i;
	char *point, *index();

	printf("FILECRC user customization...\n");
	do
		{printf("                   offset distance (0 to %ld): ", 
															0x7fffffffL);
		fgets(byte_string, 100, stdin);
		ll = atol(byte_string);
		}	while(ll < 0);
	pp->offset_distance = ll;
	printf("\nFile names may include drive letters and/or path names.\n");
	while(1)
		{
		while(1)
			{printf("    new name for this program (default %12s): ", 
															pp->crc_program);
			fgets(byte_string, 100, stdin);
			byte_string[strlen(byte_string) - 1] = 0;		/* zap the \n */
			if(byte_string[0] == 0) break;
			strupr(byte_string);
			point = index(byte_string, '.');
			if(strlen(byte_string)<sizeof(pp->crc_program) && 
												point && streq(point, ".EXE")) 
				{strcpy(pp->crc_program, byte_string);	
				break;
				}
			}
		while(1)
			{printf("  name for comparison program (default %12s): ", 
														pp->compare_program);
			fgets(byte_string, 100, stdin);
			byte_string[strlen(byte_string) - 1] = 0;		/* zap the \n */
			strupr(byte_string);
			if(byte_string[0] == 0)
				break;
			point = index(byte_string, '.');
			if(strlen(byte_string)<sizeof(pp->compare_program) && 
											point && streq(point, ".EXE")) 
				{strcpy(pp->compare_program, byte_string);	
				break;
				}
			}
		if(strcmp(pp->crc_program, pp->compare_program)) break;
		printf("CRC and comparison program pathnames must be different.\n");
		} 

	while(1)
		{while(1)
			{printf("        name for new CRC file (default %12s): ", 
														pp->new_crc_file);
			fgets(byte_string, 100, stdin);
			byte_string[strlen(byte_string) - 1] = 0;		/* zap the \n */
			if(byte_string[0] == 0) break;
			strupr(byte_string);
			if(strlen(byte_string) >= sizeof(pp->new_crc_file)) continue;
			strcpy(pp->new_crc_file, byte_string);	
			break;
			}
		while(1)
			{printf("    name for current CRC file (default %12s): ", 
														pp->cur_crc_file);
			fgets(byte_string, 100, stdin);
			byte_string[strlen(byte_string) - 1] = 0;		/* zap the \n */
			strupr(byte_string);
			if(streq(byte_string, pp->new_crc_file)) continue;
			if(strlen(byte_string) >= sizeof(pp->cur_crc_file)) continue;
			if(byte_string[0] == 0 && !streq(pp->cur_crc_file, pp->new_crc_file)) 
				break;
			strcpy(pp->cur_crc_file, byte_string);	
			break;
			}
		if(strcmp(pp->cur_crc_file, pp->new_crc_file)) break;
		printf("new and current CRC file names must be different.\n");
		}

	printf("\nThe customizable values are now...\n");
	printf("\t\t          offset: %ld\n", pp->offset_distance);
	printf("\t\t     crc program: %s\n", pp->crc_program);
	printf("\t\t compare program: %s\n", pp->compare_program);
	printf("\t\t    new CRC file: %s\n", pp->new_crc_file);
	printf("\t\tcurrent CRC file: %s\n", pp->cur_crc_file);
	printf("\nrecord these values? ");
	if(tolower(getchar()) != 'y') exit();
	printf("\n");

	srand(pp->offset_distance);
	for (i=0; i<7; i++)				/* change the search strings */
		{
		pp->flag1[ rand()%sizeof(pp->flag1) - 1 ] ^= (rand()&0x0f);
		pp->flag2[ rand()%sizeof(pp->flag2) - 1 ] ^= (rand()&0x0f);
		}
}	

/*	convert all characters in a string to upper case */	
strupr(s) char *s;
{	while(*s)
		{*s = toupper(*s);
		s++;
		}
}
